package otto

const (
	// Common classes.
	classStringName   = "String"
	classGoArrayName  = "GoArray"
	classGoSliceName  = "GoSlice"
	classNumberName   = "Number"
	classDateName     = "Date"
	classArrayName    = "Array"
	classFunctionName = "Function"
	classObjectName   = "Object"
	classRegExpName   = "RegExp"
	classBooleanName  = "Boolean"
	classMathName     = "Math"
	classJSONName     = "JSON"

	// Error classes.
	classErrorName          = "Error"
	classEvalErrorName      = "EvalError"
	classTypeErrorName      = "TypeError"
	classRangeErrorName     = "RangeError"
	classReferenceErrorName = "ReferenceError"
	classSyntaxErrorName    = "SyntaxError"
	classURIErrorName       = "URIError"

	// Common properties.
	propertyName        = "name"
	propertyLength      = "length"
	propertyPrototype   = "prototype"
	propertyConstructor = "constructor"

	// Common methods.
	methodToString = "toString"
)
